from .units import amu
#***************
#constants in SI
#***************
mp = 1.00727647*amu #kg ->proton mass 
mH = 1.00794*amu #kg ->hydrogen mass
G = 6.67384e-11 #(N m^2 kg^-2)Cavendish-gravitational constant
kb = 1.38064852e-23 #(J K^-1) Boltzmann constant
sigma = 5.670373e-8 #(W m^-2 K^-4)Stefan-Boltzmann constant
temp_cmb = 2.72548 #(K) CMB black-body temperature
c = 299792458. #(m/s) Speed of light
h = 6.62607015e-34 #(J Hz^-1) Planck constant
